from dragonk8s.lib.util.timeutil.rate import Limiter
import time


class PassiveRateLimiter(object):

    def try_accept(self) -> bool:
        pass

    def stop(self):
        pass

    def qps(self) -> float:
        pass


class RateLimiter(PassiveRateLimiter):

    def accept(self):
        pass

    def wait(self, timeout=0):
        pass


class TokenBucketPassiveRateLimiter(RateLimiter):

    def __init__(self, limiter: Limiter, qps: float):
        super().__init__()
        self._limiter = limiter
        self._qps = qps


class TokenBucketRateLimiter(TokenBucketPassiveRateLimiter):

    def __init__(self, qps: float, burst: int = 0, limiter: Limiter = None):
        if limiter is None:
            limiter = Limiter(qps, burst)
        super(TokenBucketRateLimiter, self).__init__(limiter, qps)

    def try_accept(self) -> bool:
        return self._limiter.allow_n(time.time(), 1)

    def stop(self):
        pass

    def qps(self) -> float:
        return self._qps

    def accept(self):
        now = time.time()
        time.sleep(self._limiter.reverve_n(now, 1).delay_from(now))

    def wait(self, timeout=0):
        return self._limiter.wait(timeout)


import json
import six
from kubernetes.dynamic.client import DynamicClient, ResourceInstance, ApiException, api_exception
from dragonk8s.lib.client.api_client import ApiClient
from dragonk8s.lib.apimeta import apigvk
from kubernetes.watch.watch import Watch, SimpleNamespace
from dragonk8s.lib.apimachinery.pkg.watch import watch
from queue import Queue


class SimpleBody(object):

    def __init__(self, data):
        self.data = data


def meta_request(func):
    """ Handles parsing response structure and translating API Exceptions """
    def inner(self, *args, **kwargs):
        serialize_response = kwargs.get('serialize', True)
        serializer = kwargs.get('serializer', ResourceInstance)
        if "watch" in kwargs and kwargs["watch"]:
            kwargs.pop("response_type")
        if 'response_type' in kwargs and kwargs['response_type'] is not None and kwargs['response_type'] != "":
            serialize_response = False
        try:
            resp = func(self, *args, **kwargs)
        except ApiException as e:
            raise api_exception(e)
        if serialize_response:
            try:
                if six.PY2:
                    return serializer(self, json.loads(resp.data))
                return serializer(self, json.loads(resp.data.decode('utf8')))
            except ValueError:
                if six.PY2:
                    return resp.data
                return resp.data.decode('utf8')
        return resp

    return inner


class CommonWatch(Watch, watch.Interface):

    def __init__(self, common_client):
        super(CommonWatch, self).__init__(return_type=None)
        self.common_client = common_client
        self.queue = Queue()

    def stop(self):
        pass

    def get_result_queue(self) -> Queue:
        return self.queue

    def unmarshal_event(self, data, return_type):
        js = json.loads(data)
        js['raw_object'] = js['object']
        # BOOKMARK event is treated the same as ERROR for a quick fix of
        # decoding exception
        # TODO: make use of the resource_version in BOOKMARK event for more
        # efficient WATCH
        if js['type'] != 'ERROR' and js['type'] != 'BOOKMARK':
            obj = js['raw_object']
            js['object'] = self.common_client.parse(obj)
            if hasattr(js['object'], 'metadata'):
                self.resource_version = js['object'].metadata.resource_version
            # For custom objects that we don't have model defined, json
            # deserialization results in dictionary
            elif (isinstance(js['object'], dict) and 'metadata' in js['object']
                  and 'resourceVersion' in js['object']['metadata']):
                self.resource_version = js['object']['metadata'][
                    'resourceVersion']
        return js


class CommonClient(DynamicClient):

    def _get_response_type_list(self, response_type):
        return "{}List".format(response_type)

    def __init__(self, cache_file=None, discoverer=None):
        client = ApiClient()

        super().__init__(client, cache_file, discoverer)
        self.response_type_map = {}
        self.init_response_type()
        apigvk.register_model_package()

    def init_response_type(self):
        for gvk in apigvk.ALL:
            self.register_response_type(api_version=gvk.group_version, kind=gvk.kind, response_type=gvk.response_type)

    def get_response_type(self, api_version: str, kind: str) -> str:
        key = "{}/{}".format(api_version, kind)
        if key not in self.response_type_map:
            return ""
        rt = self.response_type_map[key]
        return rt

    def get_response_type_list(self, api_version: str, kind: str) -> str:
        key = "{}/{}".format(api_version, kind)
        if key not in self.response_type_map:
            return ""
        rt = self.response_type_map[key]
        return self._get_response_type_list(rt)

    def register_response_type(self,  api_version: str, kind: str, response_type: str):
        key = "{}/{}".format(api_version, kind)
        self.response_type_map[key] = response_type

    def serialize_body(self, body):
        return body

    def ensure_namespace(self, resource, namespace, body):
        if not namespace:
            if hasattr(body, "metadata"):
                namespace = body.metadata.namespace
        if not namespace:
            raise ValueError("Namespace is required for {}.{}".format(resource.group_version, resource.kind))
        return namespace

    def list(self, resource, namespace=None, **kwargs):
        path = resource.path(name=None, namespace=namespace)
        return self.request('get', path, response_type=self.get_response_type_list(resource.group_version, resource.kind), **kwargs)

    def get(self, resource, name=None, namespace=None, **kwargs):
        return super(CommonClient, self).get(
            resource, name=name, namespace=namespace, response_type=self.get_response_type(resource.group_version, resource.kind), **kwargs)

    def create(self, resource, body=None, namespace=None, **kwargs):
        return super(CommonClient, self).create(
            resource, body=body, namespace=namespace, response_type=self.get_response_type(resource.group_version, resource.kind), **kwargs)

    def delete(self, resource, name=None, namespace=None, body=None, label_selector=None, field_selector=None, **kwargs):
        return super(CommonClient, self).delete(resource, name=name, namespace=namespace, body=body,
                                                label_selector=label_selector, field_selector=field_selector,
                                                response_type=self.get_response_type(resource.group_version, resource.kind), **kwargs)

    def replace(self, resource, body=None, name=None, namespace=None, **kwargs):
        return super(CommonClient, self).replace(resource, body=body, name=name, namespace=namespace,
                                                 response_type=self.get_response_type(resource.group_version, resource.kind), **kwargs)

    def patch(self, resource, body=None, name=None, namespace=None, **kwargs):
        return super(CommonClient, self).patch(resource, body=body, name=name, namespace=namespace,
                                               response_type=self.get_response_type(resource.group_version, resource.kind), **kwargs)

    def watch(self, resource, namespace=None, name=None, label_selector=None, field_selector=None,
              resource_version=None, timeout_seconds=None, watcher=None):
        """
        Stream events for a resource from the Kubernetes API

        :param resource: The API resource object that will be used to query the API
        :param namespace: The namespace to query
        :param name: The name of the resource instance to query
        :param label_selector: The label selector with which to filter results
        :param field_selector: The field selector with which to filter results
        :param resource_version: The version with which to filter results. Only events with
                                 a resource_version greater than this value will be returned
        :param timeout_seconds: The amount of time in seconds to wait before terminating the stream
        :param watcher: The Watcher object that will be used to stream the resource

        :return: Event object with these keys:
                   'type': The type of event such as "ADDED", "DELETED", etc.
                   'raw_object': a dict representing the watched object.
                   'object': A ResourceInstance wrapping raw_object.

        Example:
            client = DynamicClient(k8s_client)
            watcher = watch.Watch()
            v1_pods = client.resources.get(api_version='v1', kind='Pod')

            for e in v1_pods.watch(resource_version=0, namespace=default, timeout=5, watcher=watcher):
                print(e['type'])
                print(e['object'].metadata)
                # If you want to gracefully stop the stream watcher
                watcher.stop()
        """
        if not watcher: watcher = CommonWatch(self)
        cnt = 0
        for event in watcher.stream(
            resource.get,
            namespace=namespace,
            name=name,
            field_selector=field_selector,
            label_selector=label_selector,
            resource_version=resource_version,
            serialize=False,
            timeout_seconds=timeout_seconds,
            watch=True,
        ):
            cnt += 1
            watcher.get_result_queue().put(event)
        if cnt == 0:
            return None
        watcher.get_result_queue().put({
            "type": watch.EventType.End
        })
        return watcher

    @meta_request
    def request(self, method, path, body=None, **params):
        if not path.startswith('/'):
            path = '/' + path

        path_params = params.get('path_params', {})
        query_params = params.get('query_params', [])
        if params.get('pretty') is not None:
            query_params.append(('pretty', params['pretty']))
        if params.get('_continue') is not None:
            query_params.append(('continue', params['_continue']))
        if params.get('include_uninitialized') is not None:
            query_params.append(('includeUninitialized', params['include_uninitialized']))
        if params.get('field_selector') is not None:
            query_params.append(('fieldSelector', params['field_selector']))
        if params.get('label_selector') is not None:
            query_params.append(('labelSelector', params['label_selector']))
        if params.get('limit') is not None:
            query_params.append(('limit', params['limit']))
        if params.get('resource_version') is not None:
            query_params.append(('resourceVersion', params['resource_version']))
        if params.get('timeout_seconds') is not None:
            query_params.append(('timeoutSeconds', params['timeout_seconds']))
        if params.get('watch') is not None:
            query_params.append(('watch', params['watch']))
        if params.get('grace_period_seconds') is not None:
            query_params.append(('gracePeriodSeconds', params['grace_period_seconds']))
        if params.get('propagation_policy') is not None:
            query_params.append(('propagationPolicy', params['propagation_policy']))
        if params.get('orphan_dependents') is not None:
            query_params.append(('orphanDependents', params['orphan_dependents']))
        if params.get('dry_run') is not None:
            query_params.append(('dryRun', params['dry_run']))
        if params.get('field_manager') is not None:
            query_params.append(('fieldManager', params['field_manager']))
        if params.get('force_conflicts') is not None:
            query_params.append(('force', params['force_conflicts']))
        if params.get('allow_watch_bookmarks') is not None:
            query_params.append(('allowWatchBookmarks', params['allow_watch_bookmarks']))

        header_params = params.get('header_params', {})
        form_params = []
        local_var_files = {}

        # Checking Accept header.
        new_header_params = dict((key.lower(), value) for key, value in header_params.items())
        if not 'accept' in new_header_params:
            header_params['Accept'] = self.client.select_header_accept([
                'application/json',
                'application/yaml',
            ])

        # HTTP header `Content-Type`
        if params.get('content_type'):
            header_params['Content-Type'] = params['content_type']
        else:
            header_params['Content-Type'] = self.client.select_header_content_type(['*/*'])

        # Authentication setting
        auth_settings = ['BearerToken']
        preload_content = 'response_type' in params and params['response_type'] is not None and params['response_type'] != ""

        api_response = self.client.call_api(
            path,
            method.upper(),
            path_params,
            query_params,
            header_params,
            body=body,
            post_params=form_params,
            async_req=params.get('async_req'),
            files=local_var_files,
            auth_settings=auth_settings,
            _preload_content=preload_content,
            _return_http_data_only=params.get('_return_http_data_only', True),
            _request_timeout=params.get('_request_timeout'),
            response_type=params.get('response_type', None)

        )
        if params.get('async_req'):
            return api_response.get()
        else:
            return api_response

    def parse(self, resource, data):
        if not isinstance(data, dict):
            data = data.to_dict()
        return self.client.deserialize(SimpleBody(json.dumps(data)),
                                       self.get_response_type(resource.group_version, resource.kind))



from __future__ import annotations

import copy
import json
import logging
import threading
import time

from dragonk8s.lib.client.tools import cache
from dragonk8s.lib.client.tools import fifo

class DeltaType(object):
    Added = "Added"
    Updated = "Updated"
    Deleted = "Deleted"
    Replaced = "Replaced"
    Sync = "Sync"


class Delta(object):

    def __init__(self, type: DeltaType, obj):
        self.type = type
        self.obj = obj


class DeltaFIFOOptions(object):

    def __init__(self, known_objects: cache.KeyListerGetter, emit_delta_type_replaced: bool, key_function=None):
        if not key_function:
            key_function = cache.meta_namespace_key_func
        self.key_function = key_function
        self.known_objects = known_objects
        self.emit_delta_type_replaced = emit_delta_type_replaced


class ZoroLengthDeltasObject(Exception):

    def __init__(self):
        super(ZoroLengthDeltasObject, self).__init__("0 length Deltas object; can't get key")


def _is_deletion_dup(a: Delta, b: Delta) -> Delta | None:
    if b.type != DeltaType.Deleted or a.type != DeltaType.Deleted:
        return None
    if isinstance(b, cache.DeletedFinalStateUnknown):
        return a
    return b


def is_dup(a: Delta, b: Delta) -> Delta | None:
    out = _is_deletion_dup(a, b)
    if out is not None:
        return out
    return None


# 如果DeletedFinalStateUnknown是倒数第二个就删掉他
def dedup_deltas(deltas: list) -> list:
    if len(deltas) < 2:
        return deltas
    a = deltas[-1]
    b = deltas[-2]
    out = is_dup(a, b)
    if out is not None:
        deltas[-2] = out
        return deltas[:-1]
    return deltas


class DeltaFIFO(fifo.Queue):

    @staticmethod
    def new_delta_fifo(key_func, known_objects: cache.KeyListerGetter):
        return DeltaFIFO(DeltaFIFOOptions(known_objects, False, key_func))

    def __init__(self, opts: DeltaFIFOOptions):
        if not opts.key_function:
            opts.key_function = cache.meta_namespace_key_func
        self.items = {}
        self.queue = []
        self.key_func = opts.key_function
        self.known_objects = opts.known_objects
        self.emit_delta_type_replaced = opts.emit_delta_type_replaced
        self.lock = threading.Lock()
        self.closed = False
        self.cond = threading.Condition(lock=self.lock)
        self.populated = False
        self.initial_population_count = 0

    def queue_action_locked(self, action_type, obj):
        try:
            id = self.key_of(obj)
        except Exception as e:
            raise cache.KeyError(obj, e)
        old_deltas = self.items[id] if id in self.items else []
        new_deltas = copy.copy(old_deltas)
        new_deltas.append(Delta(action_type, obj))
        new_deltas = dedup_deltas(new_deltas)
        if len(new_deltas) > 0:
            if id not in self.items:
                self.queue.append(id)
            self.items[id] = new_deltas
            self.cond.notify_all()
        else:
            if not old_deltas:
                logging.error("Impossible dedupDeltas for id=%s, old_deltas=%s, obj=%s; ignoring"
                              % (id, json.dumps(old_deltas), json.dumps(obj)))
            logging.error("Impossible dedupDeltas for id=%s, old_deltas=%s, "
                          "obj=%s; breaking invariant by storing empty Deltas"
                          % (id, json.dumps(old_deltas), json.dumps(obj)))
            self.items[id] = new_deltas
            raise Exception("Impossible dedupDeltas for id=%s, old_deltas=%s,obj=%s; "
                            "breaking invariant by storing empty Deltas"
                            % (id, json.dumps(old_deltas), json.dumps(obj)))

    def list_locked(self) -> list:
        l = []
        for key, item in self.items.items():
            l.append(item[-1].obj)
        return l

    def close(self):
        try:
            self.lock.acquire()
            self.closed = True
            self.cond.notify_all()
        finally:
            self.lock.release()

    def is_closed(self) -> bool:
        with self.lock:
            return self.closed

    def key_of(self, obj) -> str:
        if isinstance(obj, list):
            if len(obj) == 0:
                raise cache.KeyError(obj, ZoroLengthDeltasObject())
            obj = obj[-1].obj
        if isinstance(obj, cache.DeletedFinalStateUnknown):
            return obj.key
        return self.key_func(obj)

    def add(self, obj):
        with self.lock:
            self.populated = True
            self.queue_action_locked(DeltaType.Added, obj)

    def update(self, obj):
        with self.lock:
            self.populated = True
            self.queue_action_locked(DeltaType.Updated, obj)

    def delete(self, obj):
        try:
            id = self.key_of(obj)
        except Exception as e:
            raise cache.KeyError(obj, e)
        with self.lock:
            self.populated = True
            if self.known_objects is None:
                if id not in self.items:
                    return
            else:
                try:
                    _, exists = self.known_objects.get_by_key(id)
                except Exception as e:
                    logging.error("known_objects get_by_key error: %s" % str(e))
                else:
                    if not exists and id not in self.items:
                        return
            return self.queue_action_locked(DeltaType.Deleted, obj)

    def list(self):
        with self.lock:
            return self.list_locked()

    def list_keys(self):
        with self.lock:
            l = []
            for key in self.queue:
                l.append(key)
        return l

    def get(self, obj):
        try:
            key = self.key_of(obj)
        except Exception as e:
            raise cache.KeyError(obj, e)
        return self.get_by_key(key)

    def get_by_key(self, key):
        with self.lock:
            if key in self.items:
                return copy.copy(self.items[key])
        return []

    def replace(self, data, resource_version):
        with self.lock:
            keys = set()
            action = DeltaType.Sync
            if self.emit_delta_type_replaced:
                action = DeltaType.Replaced

            for item in data:
                try:
                    key = self.key_of(item)
                except Exception as e:
                    raise cache.KeyError(item, e)
                keys.add(key)
                try:
                    self.queue_action_locked(action, item)
                except Exception as e:
                    raise Exception("couldn't enqueue object: %s", str(e))
            if self.known_objects is None:
                queued_deletions = 0
                for k, old_item in self.items.items():
                    if k in keys:
                        continue
                    deleted_obj = old_item[-1].obj
                    queued_deletions += 1
                    self.queue_action_locked(DeltaType.Deleted, cache.DeletedFinalStateUnknown(k, deleted_obj))
                if not self.populated:
                    self.populated = True
                    self.initial_population_count = len(keys) + queued_deletions
                return

            know_keys = self.known_objects.list_keys()
            queued_deletions = 0
            for k in know_keys:
                if k in keys:
                    continue
                try:
                    deleted_obj, exist = self.known_objects.get_by_key(k)
                except Exception as e:
                    deleted_obj = None
                    logging.error("Unexpected error %s during lookup of key: %s, "
                                  "placing DeleteFinalStateUnkonwn marker without object" % (str(e), k))
                else:
                    if not exist:
                        deleted_obj = None
                        logging.info("key: %s does not exist in konwn objects store, "
                                     "placing DeleteFinalStateUnkonwn marker without object" % k)
                queued_deletions += 1
                self.queue_action_locked(DeltaType.Deleted, cache.DeletedFinalStateUnknown(k, deleted_obj))
            if not self.populated:
                self.populated = True
                self.initial_population_count = len(keys) + queued_deletions

    def _sync_key_locked(self, key: str):
        try:
            obj, exists = self.known_objects.get_by_key(key)
        except Exception as e:
            logging.error("Unexpected error %s during lookup of key %s, unable to queue object for sync" % (str(e), key))
            return
        else:
            if not exists:
                logging.info("key %s does not exist in known objects store, unable to queue object for sync" % key)
                return
        try:
            id = self.key_of(obj)
        except Exception as e:
            raise cache.KeyError(obj, e)
        if len(self.items[id]) > 0:
            return
        try:
            self.queue_action_locked(DeltaType.Sync, obj)
        except Exception as e:
            raise Exception("couldn't queue object: %s" % str(e))
        return

    def resync(self):
        with self.lock:
            if not self.known_objects:
                return
            keys = self.known_objects.list_keys()
            for k in keys:
                self._sync_key_locked(k)

    def pop(self, pop_process_func):
        with self.lock:
            while True:
                while len(self.queue) == 0:
                    if self.closed:
                        raise fifo.FIFOClosedException()
                    self.cond.wait()
                id = self.queue.pop(0)
                depth = len(self.queue)
                if self.initial_population_count > 0:
                    self.initial_population_count -= 1
                if id not in self.items:
                    logging.error("Inconceivable! %s was in queue but not in items" % id)
                    continue
                item = self.items[id]
                del self.items[id]
                start = time.time()
                try:
                    pop_process_func(item)
                except fifo.RequeueException as e:
                    self._add_if_not_present(id, item)
                    raise e.e
                finally:
                    end = time.time()
                    if end - start >= 0.1 and depth > 10:
                        logging.warning("id: %s, depth: %s, slow event handlers blocking the queue" % (id, str(depth)))

                return item

    def _add_if_not_present(self, id: str, deltas: list):
        self.populated = True
        if id in self.items:
            return
        self.queue.append(id)
        self.items[id] = deltas
        self.cond.notify_all()

    def add_if_not_present(self, obj):
        if not isinstance(obj, list):
            raise Exception("obj must be a list of delta, but got: %s", obj.__type__)
        try:
            id = self.key_of(obj)
        except Exception as e:
            raise cache.KeyError(obj, e)
        with self.lock:
            self._add_if_not_present(id, obj)

    def has_synced(self) -> bool:
        with self.lock:
            return self.populated and self.initial_population_count == 0





def list_equal(l: list, l2: list) -> bool:
    if len(l) != len(l2):
        return False
    for i in l:
        if i not in l2:
            return False
    for i in l2:
        if i not in l:
            return False

    return True


import logging
import time
from threading import Thread, Lock

class ThreadManager(object):

    def __init__(self):
        self._data = {}
        self._lock = Lock()

    def new_thread(self, target, args=(), daemon=False, generate_name=None, name="") -> Thread:
        if generate_name is not None:
            name = generate_name + str(time.time())
        if name in self._data:
            return self._data[name]
        with self._lock:
            th = Thread(target=target, args=args, name=name)
            if daemon:
                th.setDaemon(True)
            else:
                self._data[name] = th
            return th

    def join(self):
        for name, th in self._data.items():
            logging.info("wait %s..." % name)
            th.join()

    def wait_stop_and_clean(self, timeout=30) -> bool:
        count = 0
        while count <= timeout:
            to_delete = []
            have_alive = False
            for name, th in self._data.items():
                if th.is_alive():
                    # logging.debug("thread %s is already alive" % name)
                    have_alive = True
                else:
                    to_delete.append(name)
            if not have_alive:
                self._data.clear()
                break
            for n in to_delete:
                del self._data[n]
            time.sleep(1)
            count += 1
        if len(self._data) > 0:
            logging.warning("[%s] is still in running" % ", ".join(self._data.keys()))
            return False
        return True


GlobalThreadManager  = ThreadManager()




from __future__ import annotations

import copy
import json
import logging

from dragonk8s.dragon.models import IoDragonAppsV1ReplicaSet, V1Pod, IoDragonAppsV1ReplicaSetStatus, \
    IoDragonAppsV1ReplicaSetStatusConditions, IoDragonAppsV1ReplicaSetSpec, IoDragonAppsV1JobSpecTemplateSpec
from dragonk8s.lib.apimachinery.pkg import labels
from dragonk8s.pkg.api.v1 import pod_util
import time
from dragonk8s.pkg.api import types
from dragonk8s.lib.util import timeutil


class ReplicaSetConditionType(object):
    ReplicaSetReplicaFailure = "ReplicaFailure"


def get_condition(status: IoDragonAppsV1ReplicaSetStatus, cond_type: str) -> IoDragonAppsV1ReplicaSetStatusConditions|None:
    if not status:
        return None
    if not status.conditions:
        return None
    for c in status.conditions:
        if c.type == cond_type:
            return c
    return None


def new_replication_condition(cond_type, status, reason, msg) -> IoDragonAppsV1ReplicaSetStatusConditions:
    return IoDragonAppsV1ReplicaSetStatusConditions(
        type=cond_type,
        status=status,
        last_transition_time=timeutil.to_time_str_with_ns(time.time()),
        reason=reason,
        message=msg,
    )


def _filter_out_condition(conditions, cond_type) -> list:
    new_conds = []
    for c in conditions:
        if c.type == cond_type:
            continue
        new_conds.append(c)
    return new_conds


def set_condition(status: IoDragonAppsV1ReplicaSetStatus, condition: IoDragonAppsV1ReplicaSetStatusConditions):
    current_cond = get_condition(status, condition.type)
    if current_cond is not None and current_cond.status == condition.status and current_cond.reason == condition.reason:
        return
    new_cond = _filter_out_condition(status.conditions, condition.type)
    new_cond.append(condition)
    status.conditions = new_cond


def remove_condition(status: IoDragonAppsV1ReplicaSetStatus, cond_type):
    status.conditions = _filter_out_condition(status.conditions, cond_type)


def calculate_status(rs: IoDragonAppsV1ReplicaSet,
                     filtered_pods: list, manage_replicas_exception: Exception) -> IoDragonAppsV1ReplicaSetStatus:
    new_status = copy.deepcopy(rs.status)
    if new_status is None:
        new_status = IoDragonAppsV1ReplicaSetStatus()
    fully_labeled_replicas_count = 0
    ready_replicas_count = 0
    available_replicas_count = 0
    template_label = labels.selector_from_validated_set(labels.Set(rs.spec.template.metadata.labels))
    for pod in filtered_pods:
        if template_label.matches(labels.Set(pod.metadata.labels)):
            fully_labeled_replicas_count += 1
        if pod_util.is_pod_ready(pod):
            ready_replicas_count += 1
            if pod_util.is_pod_available(pod, rs.spec.min_ready_seconds, time.time()):
                available_replicas_count += 1
    failure_cond = get_condition(rs.status, ReplicaSetConditionType.ReplicaSetReplicaFailure)
    if manage_replicas_exception is not None and failure_cond is not None:
        reason = ""
        diff = len(filtered_pods) - int(rs.status.replicas)
        if diff < 0:
            reason = "FailedCreate"
        else:
            reason = "FailedDelete"
        cond = new_replication_condition(ReplicaSetConditionType.ReplicaSetReplicaFailure,
                                         types.ConditionTrue, reason, str(manage_replicas_exception))

        set_condition(new_status, cond)
    elif manage_replicas_exception is None and failure_cond is not None:
        remove_condition(new_status, ReplicaSetConditionType.ReplicaSetReplicaFailure)

    new_status.replicas = len(filtered_pods)
    new_status.fully_labeled_replicas = fully_labeled_replicas_count
    new_status.ready_replicas = ready_replicas_count
    new_status.available_replicas = available_replicas_count
    return new_status


def update_replicaset_status(client, rs: IoDragonAppsV1ReplicaSet, new_status: IoDragonAppsV1ReplicaSetStatus) -> IoDragonAppsV1ReplicaSet:
    if rs.status is not None and new_status is not None and rs.status.replicas == new_status.replicas and \
            rs.status.fully_labeled_replicas == new_status.fully_labeled_replicas and \
            rs.status.ready_replicas == new_status.ready_replicas and \
            rs.status.available_replicas == new_status.available_replicas and \
            rs.metadata.generation == rs.status.observed_generation and \
            rs.status.conditions == new_status.conditions:
        logging.debug("no status changed for rs %s/%s" % (rs.metadata.namespace, rs.metadata.name))
        return rs
    if new_status is None:
        new_status = IoDragonAppsV1ReplicaSetStatus()
    new_status.observed_generation = rs.metadata.generation
    logging.info("Updating status for %s: %s/%s, status: %s"
                 % (rs.kind, rs.metadata.namespace, rs.metadata.name, json.dumps(new_status.to_dict())))
    rs.status = new_status
    return client.status.replace(name=rs.metadata.name, namespace=rs.metadata.namespace, body=rs)



from threading import Event
from dragonk8s.lib.apimachinery.pkg.apis import meta_v1
from dragonk8s.lib.apimachinery.pkg.apis import meta_internal_version
from kubernetes.client.exceptions import ApiException
from dragonk8s.lib.apimachinery.pkg.api import errors, meta


def simple_page_func(fn):
    def do(stop: Event, options:meta_v1.ListOptions):
        return fn(options)
    return do


class ListPager(object):

    def __init__(self, page_fn, page_size=500, full_list_if_expired=True, page_buffer_size=10):
        self.page_size = page_size
        self.page_fn = page_fn
        self.full_list_if_expired = full_list_if_expired
        self.page_buffer_size = page_buffer_size

    def list(self, options:meta_v1.ListOptions, stop: Event):
        if options.limit == 0:
            options.limit = self.page_size

        requested_resource_version = options.resource_version
        requested_resource_version_match = options.resource_version_match
        paginated_result = False
        rlist = None

        while True:
            if stop.is_set():
                return None, paginated_result
            try:
                obj = self.page_fn(stop, options)
            except ApiException as e:
                if not errors.is_resource_expired(e) or not self.full_list_if_expired or options._continue == "":
                    return None, paginated_result
                options.limit = 0
                options._continue = ""
                options.resource_version = requested_resource_version
                options.resource_version_match = requested_resource_version_match
                return self.page_fn(stop, options), paginated_result
            except Exception as e:
                return None, paginated_result
            try:
                m = meta.list_accessor(obj)
            except Exception as e:
                raise Exception("returned object is not standard list: " + str(e))
            if not m._continue and rlist is None:
                return obj, paginated_result
            options._continue = m._continue
            options.resource_version = ""
            options.resource_version_match = ""
            paginated_result = True



from dragonk8s.lib.apimachinery.pkg.apis import meta_v1


class List(object):

    def __init__(self, type_meta, list_meta, items):
        self.type_meta = type_meta
        self.list_meta = list_meta
        self.items = items


from kubernetes.client.models.v1_object_reference import V1ObjectReference
from dragonk8s.lib.apimachinery.pkg.runtime import schema


def get_reference(obj) -> V1ObjectReference:
    if obj is None:
        raise Exception("obj is None")
    if isinstance(obj, V1ObjectReference):
        return obj
    api_version = getattr(obj, "api_version")
    kind = getattr(obj, "kind")
    gvk = schema.from_api_version_and_kind(api_version, kind)
    if gvk.empty():
        raise Exception("unexecpeced gvks from obj: %s" % obj)
    object_meta = getattr(obj, "metadata")
    if object_meta is None:
        return V1ObjectReference(kind=kind, api_version=api_version)
    return V1ObjectReference(
        kind=kind,
        api_version=api_version,
        name=object_meta.name,
        namespace=object_meta.namespace,
        uid=object_meta.uid,
        resource_version=object_meta.resource_version,
    )



import time
import unittest
import logging
from threading import Event
from kubernetes import config
from dragonk8s.lib.client.tools import events
from kubernetes.client.models.v1_pod import V1Pod
from kubernetes.client.models.v1_object_meta import V1ObjectMeta
from dragonk8s.lib.client.kubernetes.typed.core.v1.event_expansion import EventSinkImpl


class EventsThreadSafeMap(unittest.TestCase):

    def setUp(self):
        pass



import json
import threading
import time
import unittest
from kubernetes import config, dynamic
from kubernetes.client import models, configuration, api_client
from dragonk8s.dragon.models import *
from dragonk8s.dragon.models import IoDragonAppsV1ReplicaSetList, IoDragonAppsV1ReplicaSetStatus
from dragonk8s.lib.client import CommonClient, CommonWatch
from dragonk8s.dragon.api_client import ApiClient
from dragonk8s.lib.client.configuration import Configuration
from kubernetes.client.models import V1ReplicaSet
from kubernetes.client.apis import CoreV1Api
from kubernetes.client.api import CoreApi
from kubernetes.client.models import V1ObjectMeta
from kubernetes import watch
"""
1. 在dragon.models.__init__.py中加上
from kubernetes.client.models import *
2. 在dragon.configuration.py中修改self.client_side_validation=false
"""


class DragonClientTest(unittest.TestCase):

    def setUp(self):
        cfg = Configuration()
        # config.load_kube_config(config_file="C:\\Users\\Administrator\\.kube\\config", client_configuration=cfg)
        config.load_kube_config(config_file="/Users/bytedance/mygit/deploy/0_config/my-k8s.conf", client_configuration=cfg)
        Configuration.set_default(cfg)
        self.base_client = ApiClient()
        # client = dynamic.DynamicClient(
        #     self.base_client
        # )
        # config.load_kube_config(config_file="/Users/bytedance/.kube/config")
        # self.base_client = api_client.ApiClient()
        try:
            client = CommonClient()
            # client2 = dynamic.DynamicClient(self.base_client)
            self.client = client.resources.get(api_version="apps.dragon.io/v1", kind="ReplicaSet")
            # self.client2 = client2.resources.get(api_version="apps.dragon.io/v1", kind="ReplicaSet")

        except EOFError as e:
            print(e)

    def test_create(self):
        expressions = IoDragonAppsV1JobSpecSelectorMatchExpressions(
                        key="key",
                        operator="in",
                        values=["v1"]
                    )
        rs = IoDragonAppsV1ReplicaSet(
            api_version="apps.dragon.io/v1",
            kind="ReplicaSet",
            metadata=models.v1_object_meta.V1ObjectMeta(
                namespace="default",
                name="test30"
            ),
            spec=IoDragonAppsV1ReplicaSetSpec(
                min_ready_seconds=111,
                replicas=1,
                selector=IoDragonAppsV1JobSpecSelector(
                    match_expressions=[expressions]
                )
            )
        )
        rss = self.client.create(rs)
        print(rss.metadata)
        print(rss.api_version)
        # print(self.base_client.sanitize_for_serialization(rs))
        # self.client.create(self.base_client.sanitize_for_serialization(rs))

    def test_get(self):
        res = self.client.get(name="tt1", namespace="default")
        print(res)

    def test_patch(self):
        data = {
            "metadata": {
                "labels": {"app": "test", "aa2": "bb2"}
            }
        }
        res = self.client.patch(body=data, name="test30", namespace="default", content_type="application/merge-patch+json")
        print(res)

    def test_list(self):
        try:
            res = self.client.list()
        except Exception as e:
            print(str(e))
            return
        print(res)
        for r in res.items:
            print("{}: {}".format(r.metadata.name, res.metadata.resource_version))
            print(res.api_version)

    def test_subresource(self):
        res = self.client.get(name="test19", namespace="default")
        status = IoDragonAppsV1ReplicaSetStatus(
            available_replicas=1,
            replicas=0,
            ready_replicas=2,
            fully_labeled_replicas=0,
            conditions=[],
        )
        res.status = status
        rs = self.client.status.replace(name="test19", namespace="default", body=res)
        print(rs)


    """
    超过最新的报
    HTTP response body: {"kind":"Status","apiVersion":"v1","metadata":{},"status":"Failure","message":"Timeout: Too large resource version: 12313646, current: 12313645","reason":"Timeout","details":{"causes":[{"reason":"ResourceVersionTooLarge","message":"Too large resource version"}],"retryAfterSeconds":1},"code":504}
    """
    def test_list_by_resource_version(self):
        res = self.client.list(namespace="default", resource_version="0")
        print(res.metadata.resource_version)
        for r in res.items:
            print("{}: {}".format(r.metadata.name, res.metadata.resource_version))

    def test_watch(self):
        rs = "0"
        w = watch.Watch()
        while True:
            for e in self.client2.watch(resource_version=rs, namespace="default", timeout=5, watcher=w):
                print("------------------------------------------")
                if e['object'] is not None:
                    rs = e['object'].metadata.resourceVersion
                    print("===")
                    print(self.client.parse(e['object']))

    def test_watch2(self):
        rs = "0"
        w = CommonWatch(common_client=self.client)
        while True:
            print("+++------------------------------------------")
            for e in self.client.watch(resource_version=rs, namespace="default", timeout_seconds=5, watcher=w):
                print("------------------------------------------")
                if e['object'] is not None:
                    rs = e['object'].metadata.resource_version
                    print("===")
                    print(e['object'])

    def test_delete(self):
        res = self.client.delete(name="test29", namespace="default")
        print(res)


    def test(self):
        res = CommonClient(self.base_client).resources.get(api_version="apps/v1", kind="ControllerRevision").list()
        for r in res.items:
            print("{}".format(r))


from threading import Lock
import time


Inf = float("inf")


def every(interval: int) -> float:
    if interval <= 0:
        return Inf
    return 1 / float(interval)


def duration_from_tokens(limit, tokens: float):
    if limit <= 0:
        return Inf
    return float(tokens/limit)


def tokens_from_duration(limit, d):
    if limit <= 0:
        return 0
    return float(d * limit)


class RateLimitException(Exception):
    pass


class Reservation(object):

    def __init__(self, ok=False, lim=None, tokens=0, time_to_act=0, limit=0):
        self._ok = ok
        self._lim = lim
        self._tokens = tokens
        self._time_to_act = time_to_act
        self._limit = limit

    @property
    def ok(self):
        return self._ok

    def delay_from(self, now):
        if not self._ok:
            return Inf
        delay = self._time_to_act - now
        if delay < 0:
            return 0
        return delay

    def delay(self):
        return self.delay_from(time.time())

    def cancel_at(self, now):
        if not self._ok:
            return
        try:
            self._lim.lock.acquire(True)
            if self._lim._limit == Inf or self._tokens == 0 or self._time_to_act < now:
                return
            restore_tokens = self._tokens - duration_from_tokens(self._limit, self._lim.last_event - self._time_to_act)
            if restore_tokens <= 0:
                return
            now, _, tokens = self._lim._advance(now)
            tokens += restore_tokens
            burst = float(self._lim._burst)
            if tokens > burst:
                tokens = burst
            self._lim._last = now
            self._lim.tokens = tokens
            if self._time_to_act == self._lim.last_event:
                prev_event = self._time_to_act + duration_from_tokens(self._limit, float(-1 * self._tokens))
                if not prev_event < now:
                    self._lim.last_event = prev_event
        finally:
            self._lim.lock.release()

    def cancel(self):
        self.cancel_at(time.time())


class Limiter(object):

    # 令牌桶, 每秒limit个, 允许b个突发
    def __init__(self, limit: float, burst: int):
        self.lock = Lock()
        self._limit = limit
        self._burst = burst
        self._tokens = float(0)
        self._last = float(0)
        self.last_event = float(0)

    @property
    def limit(self):
        try:
            self.lock.acquire(True)
            return self._limit
        finally:
            self.lock.release()

    @property
    def burst(self):
        try:
            self.lock.acquire(True)
            return self._burst
        finally:
            self.lock.release()

    # 向前推进
    def _advance(self, now):
        last = self._last
        if now < last:
            last = now
        # 比last过去了的时间
        elapsed = now - last
        # 这些时间应该产生多少令牌
        delta = tokens_from_duration(self._limit, elapsed)
        tokens = self._tokens + delta
        burst = float(self._burst)
        if tokens > burst:
            tokens = burst
        return now, last, tokens

    def _reserve_n(self, now, n: int, max_future_reserve: float) -> Reservation:
        try:
            self.lock.acquire(True)
            if self._limit == Inf:
                return Reservation(ok=True, lim=self, tokens=n, time_to_act=now)
            elif self._limit == 0:
                ok = False
                if self._burst >= n:
                    ok = True
                    self._burst -= n
                return Reservation(ok=ok, lim=self, tokens=self._burst, time_to_act=now)
            now, last, tokens = self._advance(now)
            tokens -= float(n)
            wait_duration = 0
            if tokens < 0:
                wait_duration = duration_from_tokens(self._limit, -1 * tokens)
            ok = n <= self._burst and wait_duration <= max_future_reserve
            r = Reservation(ok=ok, lim=self, limit=self._limit)
            if ok:
                r._tokens = n
                r._time_to_act = now + wait_duration

            if ok:
                self._last = now
                self._tokens = tokens
                self.last_event = r._time_to_act
            else:
                self._last = last
            return r
        finally:
            self.lock.release()

    # 申请n个令牌，返回true申请成功，可以act，返回false，申请失败，不可以act
    def allow_n(self, now, n) -> bool:
        return self._reserve_n(now, n, float(0)).ok

    def allow(self) -> bool:
        return self.allow_n(time.time(), 1)

    # 申请n个令牌, 返回一个Reservation 记为r，如果r.ok == false, 那么不允许act，因为超过了burst
    # 如果r.ok == true， 可以time.sleep(r.delay()) 就可以act，不act的话就得r.cancel()
    def reverve_n(self, now, n) -> Reservation:
        return self._reserve_n(now, n, Inf)

    def reserve(self):
        return self.reverve_n(time.time(), 1)

    # 比reverve_n的不同就是自动sleep了
    def wait_n(self, n, timeout=0):
        self.lock.acquire(True)
        burst = self._burst
        limit = self._limit
        self.lock.release()

        if n > burst and limit != Inf:
            raise RateLimitException("rate: Wait(n=%d) exceeds limiter's burst %d" % (n, burst))

        now = time.time()
        wait_limit = Inf
        if timeout > 0:
            wait_limit = timeout
        r = self._reserve_n(now, n, wait_limit)
        if not r.ok:
            raise RateLimitException("rate: Wait(n=%d) would exceed timeout" % n)

        delay = r.delay_from(now)
        if delay == 0:
            return
        time.sleep(delay)

    def wait(self, timeout=0):
        return self.wait_n(1, timeout=timeout)

    def set_limit_at(self, now, new_limit):
        try:
            self.lock.acquire(True)
            now, _, tokens = self._advance(now)
            self._last = now
            self._tokens = tokens
            self._limit = new_limit
        finally:
            self.lock.release()

    def set_limit(self, new_limit):
        self.set_limit_at(time.time(), new_limit)

    def set_burst_at(self, now, new_burst):
        try:
            self.lock.acquire(True)
            now, _, tokens = self._advance(now)
            self._last = now
            self._tokens = tokens
            self._burst = new_burst
        finally:
            self.lock.release()

    def set_burst(self, new_burst):
        self.set_burst_at(time.time(), new_burst)



import json
import threading
import time
import unittest
from kubernetes import config, dynamic
from kubernetes.client import models, configuration, api_client
from dragonk8s.dragon.models import *
from dragonk8s.dragon.models import IoDragonAppsV1ReplicaSetList
from kubernetes.client import V1ListMeta
from dragonk8s.lib.client import CommonClient, CommonWatch
from dragonk8s.dragon.api_client import ApiClient
from dragonk8s.dragon.configuration import Configuration
from kubernetes.client.models import V1ReplicaSet
from kubernetes.client.apis import CoreV1Api
from kubernetes.client.api import CoreApi
from kubernetes.client.models import V1ObjectMeta
from kubernetes import watch
from dragonk8s.lib.client.informers.dragon import replicaset
from dragonk8s.lib.client.tools.informer import ResourceEventHandler
from dragonk8s.lib.client.informers import factory
from dragonk8s.lib.apimeta import apigvk
from dragonk8s.lib.apimachinery.pkg import labels
from dragonk8s.lib.client.informers.informer import get_informer


class MyHandler(ResourceEventHandler):
    def on_add(self, obj):
        print("add %s" % obj.metadata.name)

    def on_update(self, old_obj, new_obj):
        print("update %s" % new_obj.metadata.name)

    def on_delete(self, obj):
        print("delete %s" % obj.metadata.name)


class DragonInformerTest(unittest.TestCase):

    def setUp(self):
        cfg = Configuration()
        config.load_kube_config(config_file="C:\\Users\\Administrator\\.kube\\config", client_configuration=cfg)
        Configuration.set_default(cfg)
        self.base_client = ApiClient()
        try:
            self.client = CommonClient(self.base_client)

        except EOFError as e:
            print(e)

    def test_infromer(self):
        informer = replicaset.new_replica_set_informer(self.client, "default", 0, {})
        informer.add_event_handler_with_resync_period(MyHandler(), 0)
        stop = threading.Event()
        t = threading.Thread(target=informer.run, args=(stop,))
        t.start()
        n = 0
        sync = False
        while n < 20:
            if informer.has_synced():
                print("synced------")
                sync = True
                break
            else:
                n += 1
                time.sleep(1)
        if not sync:
            raise Exception("not sync")

        t.join()

    def test_infromer2(self):
        informer = factory.new_informer(self.client, apigvk.IoDragonAppsV1ReplicaSet,  "default", 0, {}, threading.Event())
        informer.add_event_handler_with_resync_period(MyHandler(), 0)
        stop = threading.Event()
        t = threading.Thread(target=informer.run, args=(stop,))
        t.start()
        n = 0
        sync = False
        while n < 20:
            if informer.has_synced():
                print("synced------")
                sync = True
                break
            else:
                n += 1
                time.sleep(1)
        if not sync:
            raise Exception("not sync")

        t.join()

    def test_infromer3(self):
        f = factory.SharedInformerFactory(self.client, default_resync=0)


        informer = f.informer_for(apigvk.IoDragonAppsV1ReplicaSet)
        informer.add_event_handler_with_resync_period(MyHandler(), 0)
        stop = threading.Event()
        f.start(stop)
        f.wait_for_cache_sync(stop)
        print("wait done")
        time.sleep(100)

    def test_informer4(self):
        f = factory.SharedInformerFactory(self.client, default_resync=0)
        inf = replicaset.ReplicaSetInformer(f)
        informer = inf.informer()
        lister = inf.lister()
        informer.add_event_handler_with_resync_period(MyHandler(), 0)
        stop = threading.Event()
        f.start(stop)
        f.wait_for_cache_sync(stop)
        print("wait done")

        rss = lister.list(labels.every_thing)
        print(len(rss))
        rs = lister.with_namespace("default").get("tt1")
        print(rs)
        time.sleep(100)

    def test_informer5(self):
        f = factory.SharedInformerFactory(self.client, default_resync=0)
        inf = get_informer(apigvk.CoreV1Pod, f)
        informer = inf.informer()
        lister = inf.lister()
        informer.add_event_handler_with_resync_period(MyHandler(), 0)
        stop = threading.Event()
        f.start(stop)
        f.wait_for_cache_sync(stop)
        print("wait done")

        # rss = lister.list(labels.every_thing)
        # print(len(rss))
        # for rs in rss:
        #     print(rs)
        #     print("--------------")
        time.sleep(100)



from datetime import datetime
from pytz import utc


time_format = "%Y-%m-%dT%H:%M:%SZ"
time_format_ns = "%Y-%m-%dT%H:%M:%S.%fZ"


def _timestamp2str(t: float) -> str:
    return datetime.fromtimestamp(t).strftime(time_format)


def _datetime2str(t: datetime) -> str:
    return t.strftime(time_format)


def _timestamp2datetime(t: float) -> datetime:
    return datetime.fromtimestamp(t)


def to_time_str(t) -> str:
    if isinstance(t, str):
        return t
    if isinstance(t, float) or isinstance(t, int):
        return _timestamp2str(t)
    return _datetime2str(t)


def to_time_str_with_ns(t) -> str:
    if isinstance(t, datetime):
        t = t.timestamp()
    ns = t - float(int(t))
    tf = time_format_ns.format(ns)
    return datetime.fromtimestamp(t).astimezone(utc).strftime(tf)


def parse_time(t) -> float:
    if isinstance(t, float) or isinstance(t, int):
        return t
    if isinstance(t, str):
        return datetime.strptime(t, time_format).timestamp()
    return t.timestamp()



from __future__ import annotations

import logging
import queue
import threading
import time
from dragonk8s.lib.pool.thread_manager import GlobalThreadManager


class Interface(object):

    def add(self, item):
        pass

    def len(self) -> int:
        pass

    def get(self) -> (object, bool):
        pass

    def done(self, item):
        pass

    def shutdown(self):
        pass

    def shutdown_with_drain(self):
        pass

    def shutting_down(self):
        pass


class Type(Interface):

    def __init__(self, update_period=0.5):
        self.queue = []
        self.dirty = set()
        self.processing = set()
        self.lock = threading.Lock()
        self.cond = threading.Condition(self.lock)
        self._shutting_down = False
        self._drain = False
        self.unfinished_work_update_period = update_period

    def add(self, item):
        with self.lock:
            if self._shutting_down:
                return
            if item in self.dirty:
                return
            self.dirty.add(item)
            if item in self.processing:
                return
            self.queue.append(item)
            self.cond.notify(1)

    def len(self) -> int:
        with self.lock:
            return len(self.queue)

    def get(self) -> (object, bool):
        with self.lock:
            while len(self.queue) == 0 and not self._shutting_down:
                self.cond.wait()
            if len(self.queue) == 0:
                return None, True
            item = self.queue.pop(0)
            self.processing.add(item)
            self.dirty.remove(item)
            return item, False

    def done(self, item):
        with self.lock:
            self.processing.remove(item)
            if item in self.dirty:
                self.queue.append(item)
                self.cond.notify(1)
            elif len(self.processing) == 0:
                self.cond.notify(1)

    def set_drain(self, should_drain):
        with self.lock:
            self._drain = should_drain

    def should_drain(self):
        with self.lock:
            return self._drain

    def _shutdown(self):
        with self.lock:
            self._shutting_down = True
            self.cond.notify_all()

    def shutdown(self):
        self.set_drain(False)
        self._shutdown()

    def _is_processing(self) -> bool:
        with self.lock:
            return len(self.processing) != 0

    def _wait_for_processing(self):
        with self.lock:
            if len(self.processing) == 0:
                return
            self.cond.wait()

    def shutdown_with_drain(self):
        self.set_drain(True)
        self._shutdown()
        while self._is_processing() and self.should_drain():
            self._wait_for_processing()

    def shutting_down(self):
        with self.lock:
            return self._shutting_down


class DelayingInterface(Type):

    def add_after(self, item, duration: int):
        pass


class _WaitFor(object):

    def __init__(self, data, ready_at):
        self.data = data
        self.ready_at = ready_at
        self.index = 0


class _PriorityQueue(object):

    def __init__(self, size=1000):
        self._data = [_WaitFor(None, 0)] * size
        self._size = size
        self._len = 0

    def is_full(self):
        return self._len == self._size

    def is_empty(self):
        return self._len == 0

    def add(self, x: _WaitFor):
        if self.is_full():
            raise Exception("is full")
        i = self._len
        self._len += 1
        while i > 0:
            p = int((i-1)/2)
            if self._data[p].ready_at <= x.ready_at:
                break
            self._data[i] = self._data[p]
            i = p
        self._data[i] = x
        x.index = i

    def remove(self, x: _WaitFor):
        i = 0
        while i < self._len:
            if self._data[i].data == x.data:
                self._data[i] = self.data[self._len-1]
                self._len -= 1
                break

    def get(self) -> _WaitFor|None:
        if self.is_empty():
            return None
        res = self._data[0]
        i = 0
        x = self._data[self._len-1]
        self._len -= 1
        while 2 * i + 1 < self._len:
            a = 2 * i + 1
            b = 2 * i + 2
            if b < self._len and self._data[b].ready_at < self._data[a].ready_at:
                a = b
            if x.ready_at < self._data[a].ready_at:
                break
            self._data[i] = self._data[a]
            i = a
        self._data[i] = x
        return res

    def peek(self):
        if self.is_empty():
            return None
        return self._data[0]

    @property
    def data(self):
        return self._data[:self._len]

    def __len__(self):
        return self._len


class DelayingType(DelayingInterface):

    def __init__(self):
        self.stop = threading.Event()
        self.heartbeat = 10
        self.waiting_for_addch = queue.Queue(maxsize=1000)
        super(DelayingType, self).__init__()
        GlobalThreadManager.new_thread(target=self._waiting_loop, generate_name="DelayingType-_waiting_loop").start()

    def add_after(self, item, duration: int):
        if self.shutting_down():
            return
        if duration <=0:
            self.add(item)
            return
        wf = _WaitFor(data=item, ready_at=time.time()+duration)
        try:
            self.waiting_for_addch.put(wf, block=False)
        except queue.Full:
            logging.error("put failed in add_after")

    def shutdown(self):
        if not self.stop.is_set():
            logging.info("DelayingType shutdown...")
            super(DelayingType, self).shutdown()
            self.stop.set()
            logging.error("DelayingType shutdown done...")

    def _waiting_loop(self):
        try:
            waiting_queue = _PriorityQueue()
            waiting = {}

            while True:
                if self.shutting_down():
                    logging.info("quit _waiting_loop...")
                    return
                now = time.time()
                while len(waiting_queue) > 0:
                    entry = waiting_queue.peek()
                    if entry.ready_at > now:
                        break
                    entry = waiting_queue.get()
                    self.add(entry.data)
                    del waiting[entry.data]
                waiting_entry = None
                while waiting_entry is None:
                    if self.shutting_down():
                        break
                    try:
                        waiting_entry = self.waiting_for_addch.get(block=True, timeout=5)
                    except queue.Empty:
                        continue
                if waiting_entry is None:
                    continue
                if waiting_entry.ready_at > time.time():
                    if waiting_entry.data in waiting:
                        existing = waiting[waiting_entry.data]
                        existing.ready_at = waiting_entry.ready_at
                        waiting_queue.remove(existing)
                        waiting_queue.add(existing)
                    else:
                        waiting_queue.add(waiting_entry)
                        waiting[waiting_entry.data] = waiting_entry

                else:
                    self.add(waiting_entry.data)

        # todo
        except Exception as e:
            logging.error("_waiting_loop error: %s" % e)


class RateLimiter(object):

    def when(self, item) -> int:
        pass

    def forget(self, item):
        pass

    def num_requeues(self, item) -> int:
        pass


# todo
class SimpleRateLimiter(RateLimiter):
    def when(self, item) -> int:
        return 5

    def forget(self, item):
        pass

    def num_requeues(self, item) -> int:
        return 0


def default_rate_limiter():
    return SimpleRateLimiter()


class RateLimitingInterface(DelayingType):

    def add_rate_limited(self, item):
        pass

    def forget(self, item):
        pass

    def num_requeues(self, itme) -> int:
        pass


class RateLimitingType(RateLimitingInterface):

    def __init__(self, rate_limiter: RateLimiter=None):
        super(RateLimitingType, self).__init__()
        if not rate_limiter:
            rate_limiter = default_rate_limiter()
        self.rate_limiter = rate_limiter

    def add_rate_limited(self, item):
        self.add_after(item, self.rate_limiter.when(item))

    def forget(self, item):
        self.rate_limiter.forget(item)

    def num_requeues(self, itme) -> int:
        return self.rate_limiter.num_requeues(itme)





class FieldErrorType(object):
    ErrorTypeNotFound = "FieldValueNotFound"
    ErrorTypeRequired = "FieldValueRequired"
    ErrorTypeDuplicate = "FieldValueDuplicate"
    ErrorTypeInvalid = "FieldValueInvalid"
    ErrorTypeNotSupported = "FieldValueNotSupported"
    ErrorTypeForbidden = "FieldValueForbidden"
    ErrorTypeTooLong = "FieldValueTooLong"
    ErrorTypeTooMany = "FieldValueTooMany"
    ErrorTypeInternal = "InternalError"
    ErrorTypeTypeInvalid = "FieldValueTypeInvalid"


OmitValue = "__omit__"


class FieldError(Exception):

    def __init__(self, _type, field, bad_value=OmitValue, detail=""):
        self.type = _type
        self.field = field
        self.bad_value = bad_value
        self.detail = detail

    def error_body(self):
        res = ""
        if self.type in {FieldErrorType.ErrorTypeRequired,
                         FieldErrorType.ErrorTypeForbidden,
                         FieldErrorType.ErrorTypeTooLong,
                         FieldErrorType.ErrorTypeInternal} or self.bad_value == OmitValue:
            res = self.type
        else:
            res = "{}: {}".format(self.type, str(self.bad_value))
        if len(self.detail) > 0:
            res += ": {}".format(self.detail)
        return res

    def __repr__(self):
        return "{}:{}".format(self.field, self.error_body())

    def __str__(self):
        return self.__repr__()


class Path(object):

    def __init__(self, name="", index="", parent=None):
        self.name = name
        self.parent = parent
        self._index = index

    def root(self):
        p = self
        while p.parent is not None:
            p = p.parent
        return p

    def child(self, name, *more_names):
        r = new_path(name, *more_names)
        r.root().parent = self
        return r

    def index(self, index: int):
        return Path(index=str(index), parent=self)

    def key(self, key: str):
        return Path(index=key, parent=self)

    def __repr__(self):
        if not self:
            return "<nil>"
        elems = []
        p = self
        while p is not None:
            elems.append(p)
            p = p.parent
        res = ""
        for i in range(len(elems)):
            p = elems[len(elems)-1-i]
            if p.parent is not None and len(p.name) > 0:
                res += "."
            if len(p.name) > 0:
                res += p.name
            else:
                res += "[%s]" % p.index
        return res

    def __str__(self):
        return self.__repr__()


def new_path(name: str, *more_names) -> Path:
    r = Path(name=name)
    for another_name in more_names:
        r = Path(name=another_name, parent=r)
    return r


class _PathOptions(object):

    def __init__(self, path: Path = None):
        self.path = path


def to_path(opts) -> Path:
    c = _PathOptions()
    for opt in opts:
        opt(c)
    if c.path is None:
        c.path = Path()
    return c.path


def invalid(field: Path, value, detail) -> FieldError:
    return FieldError(FieldErrorType.ErrorTypeTypeInvalid, str(field), value, detail)


def duplicate(field: Path, value) -> FieldError:
    return FieldError(FieldErrorType.ErrorTypeDuplicate, str(field), value)


def not_supported(field: Path, value, valid_values: list) -> FieldError:
    detail = ""
    if len(valid_values) > 0:
        quoted_values = [str(v) for v in valid_values]
        detail = "supported values: " + ", ".join(quoted_values)
    return FieldError(FieldErrorType.ErrorTypeNotSupported, str(field), value, detail)



import json
from dragonk8s.lib.client.util.trans import parse_json_to_object_by_class_name
from dragonk8s.dragon.exceptions import ApiException
from kubernetes.client.models import V1StatusCause, V1Status, V1StatusDetails
from dragonk8s.lib.apimachinery.pkg.apis import meta_v1

def reason_for_error(ex) -> str:
    if ex is not None and ex.body is not None:
        try:
            status = json.loads(ex.body)
        except:
            return 'unknow'
        else:
            if 'reason' in status:
                return status['reason']
    return 'unknow'


def is_already_exists(ex) -> bool:
    return reason_for_error(ex) == 'AlreadyExists'


CauseTypeFieldValueNotFound = "FieldValueNotFound"
CauseTypeFieldValueRequired = "FieldValueRequired"
CauseTypeFieldValueDuplicate  = "FieldValueDuplicate"
CauseTypeFieldValueInvalid = "FieldValueInvalid"
CauseTypeFieldValueNotSupported = "FieldValueNotSupported"
CauseTypeUnexpectedServerResponse = "UnexpectedServerResponse"
CauseTypeFieldManagerConflict = "FieldManagerConflict"
CauseTypeResourceVersionTooLarge = "ResourceVersionTooLarge"
NamespaceTerminatingCause  = "NamespaceTerminating"


StatusReasonUnknown = ""
StatusReasonUnauthorized = "Unauthorized"
StatusReasonForbidden = "Forbidden"
StatusReasonNotFound = "NotFound"
StatusReasonAlreadyExists = "AlreadyExists"
StatusReasonConflict = "Conflict"
StatusReasonGone = "Gone"
StatusReasonInvalid = "Invalid"
StatusReasonServerTimeout = "ServerTimeout"
StatusReasonTimeout = "Timeout"
StatusReasonTooManyRequests = "TooManyRequests"
StatusReasonBadRequest = "BadRequest"
StatusReasonMethodNotAllowed = "MethodNotAllowed"
StatusReasonNotAcceptable = "NotAcceptable"
StatusReasonRequestEntityTooLarge = "RequestEntityTooLarge"
StatusReasonUnsupportedMediaType = "UnsupportedMediaType"
StatusReasonInternalError = "InternalError"
StatusReasonExpired = "Expired"
StatusReasonServiceUnavailable = "ServiceUnavailable"


known_reason = {StatusReasonUnauthorized, StatusReasonForbidden, StatusReasonNotFound, StatusReasonAlreadyExists,
                StatusReasonConflict, StatusReasonGone, StatusReasonInvalid, StatusReasonServerTimeout,
                StatusReasonTimeout, StatusReasonTooManyRequests, StatusReasonBadRequest, StatusReasonMethodNotAllowed,
                StatusReasonNotAcceptable, StatusReasonRequestEntityTooLarge, StatusReasonUnsupportedMediaType,
                StatusReasonInternalError, StatusReasonExpired, StatusReasonServiceUnavailable}


def status_cause(e: ApiException, cause_name: str) -> (V1StatusCause, bool):
    if not hasattr(e, "body"):
        return V1StatusCause(), False
    body = e.body
    status = parse_json_to_object_by_class_name(body, "V1Status")
    if not status.details or not status.details.causes:
        return V1StatusCause(), False
    for cause in status.details.causes:
        if cause.reason == cause_name:
            return cause, True
    return V1StatusCause(), False


def has_status_cause(e: ApiException, cause_name: str) -> bool:
    _, ok = status_cause(e, cause_name)
    return ok


def _reason_and_code_for_error(e: ApiException) -> (str, int):
    if not hasattr(e, "body"):
        return StatusReasonUnknown, 0
    body = e.body
    status = parse_json_to_object_by_class_name(body, "V1Status")
    return status.reason, status.code


def is_not_found(e: ApiException) -> bool:
    reason, code = _reason_and_code_for_error(e)
    if reason == StatusReasonNotFound:
        return True
    if reason in known_reason:
        return code == 404
    return False


def is_invalid(e: ApiException) -> bool:
    reason, code = _reason_and_code_for_error(e)
    if reason == StatusReasonInvalid:
        return True
    if reason not in known_reason and code == 422:
        return True
    return False

def is_resource_expired(e: ApiException) -> bool:
    return reason_for_error(e) == StatusReasonExpired


def is_timeout(e: ApiException) -> bool:
    reason, code = _reason_and_code_for_error(e)
    if reason == StatusReasonTimeout:
        return True
    if reason not in known_reason and code == 504:
        return True
    return False


def is_too_large_resource_version_error(e: ApiException) -> bool:
    if has_status_cause(e, CauseTypeResourceVersionTooLarge):
        return True
    if not is_timeout(e):
        return False
    if not hasattr(e, "body"):
        return False
    body = e.body
    status = parse_json_to_object_by_class_name(body, "V1Status")
    if not status.details or not status.details.causes:
        return False
    for cause in status.details.causes:
        if cause.message == "Too large resource version":
            return True
    return False


def is_too_many_requests(e: ApiException) -> bool:
    reason, code = _reason_and_code_for_error(e)
    if reason == StatusReasonTooManyRequests:
        return True
    if code == 429:
        return True
    return False


def new_not_found(gvk, name: str) -> ApiException:
    status = V1Status(
        status=meta_v1.StatusFailure,
        code=404,
        reason=StatusReasonNotFound,
        details=V1StatusDetails(
            group=gvk.group,
            kind=gvk.kind,
            name=name,
        ),
    )
    res = ApiException(status=meta_v1.StatusFailure, reason=StatusReasonNotFound)
    res.body = json.dumps(status.to_dict())
    return res





class PatchType(object):

    MergePatchType = "application/merge-patch+json"
    JSONPatchType = "application/json-patch+json"
    StrategicMergePatchType = "application/strategic-merge-patch+json"
    ApplyPatchType =  "application/apply-patch+yaml"